"""Tools for working with Adams Drill Databases

"""
import glob
import os

from .utilities import fullNotation_to_cdbNotation

def fix_strings_in_cdb(cdb_path, old_cdb_path=None, cdb_name=None):
    """Finds all the string files in the Adams Drill Database given in `cdb_path` and converts all cdb references in the string files to cdb notation.  If `old_cdb_path` and `cdb_name` are given, replaces any references to `old_cdb_path` with references to `cdb_name` using cdb notation.
    
    Parameters
    ----------
    cdb_path : str
        Path to an Adams Drill Database
    old_cdb_path : str
        Filepath to a cdb which should be replaced. (default is None)
    cdb_name : str
        Name of the cdb with which to replace the old file path. (default is None)

    """
    for string_file in glob.glob(os.path.join(cdb_path, '**', '*.str'), recursive=True):
        # For each string file in the database
        
        # If an old_cdb_path is given, replace it with the new cdb notation
        if old_cdb_path is not None and cdb_name is not None:
            replace_old_cdb_paths_in_string(string_file, old_cdb_path, cdb_name)

        # Convert full notation to cdb notation
        fullNotation_to_cdbNotation(string_file)

def replace_old_cdb_paths_in_string(string_file, old_cdb_path, cdb_name):
    """Replaces all occurences of `old_cdb_path` with '<`cdb_name`>' in `string_file`.
    
    Parameters
    ----------
    string_file : str
        String file
    old_cdb_path : str
        Filepath to a cdb which should be replaced.
    cdb_name : str
        Name of the cdb with which to replace the old file path.

    """
    # remove the cdb extension from cdb_name
    cdb_name = cdb_name.replace('.cdb', '')

    # Open the string file for reading and a temporary string file for writing
    with open(string_file, 'r') as fid, open(string_file + '.tmp', 'w') as fid_new:
        
        for line in fid:
            # For each line check if the line has the erroneous path and replace it
            line = line.replace(old_cdb_path.replace('/', '\\'), f'<{cdb_name}>').replace(old_cdb_path.replace('\\', '/'), f'<{cdb_name}>')            
            fid_new.write(line)
    
    # Replace the original file with the temp file
    os.remove(string_file)
    os.rename(string_file + '.tmp', string_file)



"""Tests related to the adripy.database module

"""
import unittest
import os
import shutil
from test import *

from adamspy.adripy import database


class Test_Database(unittest.TestCase):

    def setUp(self):
        # Try to delete the temporary database just in case it's already there
        try:
            shutil.rmtree(TEST_BAD_REFS_CDB_PATH + '.tmp')
        except FileNotFoundError:
            pass

        # Copy the bad refs database to a temporary database
        shutil.copytree(TEST_BAD_REFS_CDB_PATH, TEST_BAD_REFS_CDB_PATH + '.tmp')
        
    def test_replace_old_cdb_paths_in_string(self):
        """Tests that adripy.database.replace_old_cdb_paths_in_string produces the expected string file.

        """
        string_file = os.path.join(TEST_BAD_REFS_CDB_PATH + '.tmp', 'drill_strings.tbl', 'test_string.str')
        database.replace_old_cdb_paths_in_string(string_file, TEST_BAD_REF, TEST_BAD_REF_NEW_CDB)

        failures = check_file_contents(string_file, TEST_EXPECTED_STRING_AFTER_BAD_REF_REPLACE)

        self.assertListEqual(failures, [])

    def test_fix_strings_in_cdb(self):
        """Tests that adripy.database.fix_strings_in_cdb produces the expected string files.

        """
        database.fix_strings_in_cdb(TEST_BAD_REFS_CDB_PATH + '.tmp', old_cdb_path=TEST_BAD_REF, cdb_name=TEST_BAD_REF_NEW_CDB)
        
        # Check the string file contents
        string_file = os.path.join(TEST_BAD_REFS_CDB_PATH + '.tmp', 'drill_strings.tbl', 'test_string.str')
        failures = check_file_contents(string_file, TEST_EXPECTED_STRING_AFTER_BAD_REF_REPLACE)

        self.assertListEqual(failures, [])

    def tearDown(self):
        # Try to delete the temporary database
        try:
            shutil.rmtree(TEST_BAD_REFS_CDB_PATH + '.tmp')
        except FileNotFoundError:
            pass



import pandas as pd

def import_spreadsheet(filename, output_type='dict'):
    """Reads data from a file generated in Adams View using File>Export>Spreadsheet.
    
    Parameters
    ----------
    filename : str
        Name of file.
    output_type : str
        Type of object to return.  Default is 'DataFrame' which returns a pandas `:obj:DataFrame` object.  You can also select 'dict'.

    Returns
    -------
    DataFrame or dict
        Pandas `:obj:Dataframe` if `output_type` is 'DataFrame' or `:obj:dict` if `output_type` is 'dict'

    """
    data = pd.read_csv(filename, sep='\\s+', quotechar='"', index_col=0, skiprows=6)

    new_column_map = {c_nm : c_nm.split('.')[-1] for c_nm in data.columns}
    
    data.rename(columns=new_column_map, inplace=True)         
    data.index.rename(data.index.name.split('.')[-1], inplace=True)

    if output_type.lower() == 'dict':
        data.reset_index(inplace=True)
        data = data.to_dict('list')

    return data



"""This module is for sending scripts through Adams/PPT

"""
import os
import subprocess
import re
import time

import jinja2
from numpy import genfromtxt
import matplotlib.pyplot as plt
from thornpy.signal import manually_clean_sig, remove_data_point


LOG_COMPLETE_PATTERN = '! Command file is exhausted, batch run is finished.'

LUNAR_SCRIPT_NAME = '_get_lunar_results.py'
GET_RESULTS_SCRIPT_NAME = '_get_results.py'
EDIT_RESULTS_SCRIPT_NAME = '_edit_results.py'
TEMP_OUTPUT_FILENAME = '_results_.tmp'

LOG_NAME = 'aview.log'

TMPLT_ENV = jinja2.Environment(
    loader=jinja2.PackageLoader('adamspy.postprocess', 'aview_scripts'),
    autoescape=False,
    keep_trailing_newline=True,
    trim_blocks=True,
    lstrip_blocks=True
)

_TIMEOUT = 300
LOG_FILE_ERROR_PATTERN = '! \\S*Error: '

def get_results(res_file, reqs_to_get, t_min=None, t_max=None, _just_write_script=False, timeout=_TIMEOUT):
    template = TMPLT_ENV.from_string(open(os.path.join(os.path.dirname(__file__), 'aview_scripts', GET_RESULTS_SCRIPT_NAME)).read())
    working_directory = os.path.dirname(res_file)

    with open(os.path.join(working_directory, GET_RESULTS_SCRIPT_NAME), 'w') as fid:
        fid.write(template.render({'res_file': os.path.split(res_file)[-1], 'reqs_to_get': reqs_to_get, 't_min': t_min, 't_max': t_max, 'output_file': TEMP_OUTPUT_FILENAME}))

    if _just_write_script is False:
        # Delete the aview.log file
        try:
            os.remove(os.path.join(working_directory, LOG_NAME))
        except FileNotFoundError:
            pass
        except PermissionError:
            pass

        # Run the postprocessor
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW        
        subprocess.Popen('"{}" aview ru-s b {}'.format(os.environ['ADAMS_LAUNCH_COMMAND'], GET_RESULTS_SCRIPT_NAME), cwd=working_directory, startupinfo=startupinfo)

        # Wait for complete
        _wait(os.path.join(working_directory, LOG_NAME), timeout=timeout)

        # Check the log file for errors
        _get_log_errors(os.path.join(working_directory, LOG_NAME))

        # Read and return the results
        data = genfromtxt(os.path.join(working_directory, TEMP_OUTPUT_FILENAME), delimiter=',', names=True, dtype=None)

        os.remove(os.path.join(working_directory, GET_RESULTS_SCRIPT_NAME))
        os.remove(os.path.join(working_directory, TEMP_OUTPUT_FILENAME))

        output_dict = {'time': list(data['time'])}
        for res in reqs_to_get:
            output_dict[res] = {}
            for comp in reqs_to_get[res]:
                output_dict[res][comp] = list(data[f'{res}_{comp}'])
        
        return output_dict

def write_results(res_file, input_dict):
    # TODO
    return

def edit_results(res_file, input_dict, new_res_file=None, _just_write_script=False, timeout=_TIMEOUT):
    template = TMPLT_ENV.from_string(open(os.path.join(os.path.dirname(__file__), 'aview_scripts', EDIT_RESULTS_SCRIPT_NAME)).read())
    working_directory = os.path.dirname(res_file)

    new_res_file = os.path.split(res_file)[-1] if new_res_file is None else os.path.split(new_res_file)[-1]

    with open(os.path.join(working_directory, EDIT_RESULTS_SCRIPT_NAME), 'w') as fid:
        fid.write(template.render({'res_file': os.path.split(res_file)[-1], 'reqs_to_edit': input_dict, 'output_file': new_res_file}))

    if _just_write_script is False:
        # Delete the aview.log file
        try:
            os.remove(os.path.join(working_directory, LOG_NAME))
        except FileNotFoundError:
            pass
        except PermissionError:
            pass

        # Run the postprocessor
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW        
        subprocess.Popen('"{}" aview ru-s b {}'.format(os.environ['ADAMS_LAUNCH_COMMAND'], EDIT_RESULTS_SCRIPT_NAME), cwd=working_directory, startupinfo=startupinfo)

        # Wait for complete
        _wait(os.path.join(working_directory, LOG_NAME), timeout=timeout)

        # Check the log file for errors
        _get_log_errors(os.path.join(working_directory, LOG_NAME))

def get_lunar_results(res_files, reqs_to_get, t_min, t_max, output_file, _just_write_script=False, timeout=_TIMEOUT):
    
    template = TMPLT_ENV.from_string(open(os.path.join(os.path.dirname(__file__), 'aview_scripts', LUNAR_SCRIPT_NAME)).read())
    
    working_directory = os.path.dirname(res_files[0])

    with open(os.path.join(working_directory, LUNAR_SCRIPT_NAME), 'w') as fid:
        fid.write(template.render({'res_files': res_files, 'reqs_to_get': reqs_to_get, 't_min': t_min, 't_max': t_max, 'output_suffix': output_file}))

    if _just_write_script is False:
        # Delete the aview.log file
        try:
            os.remove(os.path.join(working_directory, LOG_NAME))
        except FileNotFoundError:
            pass

        # Run the postprocessor
        startupinfo = subprocess.STARTUPINFO()
        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW        
        subprocess.Popen('"{}" aview ru-s b {}'.format(os.environ['ADAMS_LAUNCH_COMMAND'], LUNAR_SCRIPT_NAME), cwd=working_directory, startupinfo=startupinfo)

        # Wait for complete
        _wait(os.path.join(working_directory, LOG_NAME), timeout=timeout)

        os.remove(os.path.join(working_directory, LUNAR_SCRIPT_NAME))

        # Make a list of the files that are written
        res_output_files = [os.path.splitext(output_file)[0] + '_time' + os.path.splitext(output_file)[-1]]
        for res_name, res_comps in reqs_to_get.items():
            for res_comp in res_comps:
                full_res_name = f'{res_name}_{res_comp}'
                res_output_files.append(os.path.splitext(output_file)[0] + '_' + full_res_name + os.path.splitext(output_file)[-1])

        return res_output_files

def _wait(log_file, sleep_time=0.2, timeout=300):
    """Waits for the log file to write the last line of the macro
    
    Parameters
    ----------
    log_file : str
        filename of log file
    sleep_time : float, optional
        Time between checks, by default 0.2
    timeout : int, optional
        During after which to time out, by default 300

    """    

    for _i in range(int(timeout/sleep_time)):
        
        ppt_log_file_exists = os.path.exists(log_file)       

        if ppt_log_file_exists is True:
            # If ppt.log exists, open it and see if the results have been loaded
            with open(log_file, 'r') as fid:
                text = fid.read()
            if re.search(LOG_COMPLETE_PATTERN, text):
                break
                
        time.sleep(sleep_time)

def _get_log_errors(log_file):
    """Checks the log file for errors of the type AviewError.
    
    Parameters
    ----------
    log_file : str
        Filename of aview log file (usulally aview.log)

    """
    with open(log_file, 'r') as fid:
        lines = fid.readlines()
    
    for line in lines:
        if re.search(LOG_FILE_ERROR_PATTERN, line):
            raise AviewError(line[2:])        

def manually_remove_spikes(res_file, reqs_to_clean, reqs_to_check=None, t_min=None, t_max=None, _just_write_script=False, timeout=_TIMEOUT, _inplace=False):
    """Allows the user to manually scan through the result sets to pick out points to eliminate.

    Parameters
    ----------
    res_file : str
        Adams Results (.res) filename 
    reqs_to_clean : dict
        Nested dictionary of result sets and result components to clean
    t_min : float, optional
        Minumum simulation time to clean, by default None
    t_max : float, optional
        Maximum simulation time to clean, by default None
    timeout : float, optional
        Number of seconds to wait for results to load before timing out, by default _TIMEOUT

    Returns
    -------
    dict
        Nested dictionary of cleaned results
        
    """
    if reqs_to_check is None:
        reqs_to_check = reqs_to_clean

    results = get_results(res_file, reqs_to_clean, t_min=t_min, t_max=t_max, _just_write_script=_just_write_script, timeout=timeout)
    
    time_sig = results['time']
    
    # Remove the spikes
    for (res, res_comps) in [(r, rc) for r, rc in results.items() if r in reqs_to_check]:
        for (res_comp, values) in [(rc, v) for rc, v in res_comps.items() if rc in reqs_to_check[res]]:
            results[res][res_comp], i_mod = manually_clean_sig(time_sig, values, indices=True)

            # If a modification was made to the signal, make that modification to the rest of the signals
            if i_mod != []:

                # Loop over all the other results
                for (other_res, other_res_comps) in [(r, rc) for r, rc in results.items() if r != 'time']:
                    for (other_res_comp, other_values) in [(rc, v) for rc, v in other_res_comps.items() if not (other_res == res and rc == res_comp)]: #pylint: disable=no-member
                        for i in i_mod:
                            results[other_res][other_res_comp] = remove_data_point(time_sig, other_values, i)

    # Update the analysis files
    edit_results(res_file, results)

    # Return the cleaned results
    return results
    
class AviewError(Exception):
    """Raise this error to if a known error occurs in the log file.
    
    """
    pass



from adamspy.postprocess.ppt import get_results as _get_results

def get_res_data(res_file, res, comp, t_min=None, t_max=None):
    results = _get_results(res_file, {res: [comp]}, t_min=t_min, t_max=t_max)    
    return results['time'], results[res][comp], len(results['time'])